import logging
import time
from logging import Logger
from threading import Thread
from typing import List

from injector import Injector

from isar.apis.api import API
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.log import setup_loggers
from isar.config.settings import settings
from isar.models.communication.queues.queues import Queues
from isar.modules import get_injector
from isar.services.service_connections.mqtt.mqtt_client import MqttClient
from isar.services.service_connections.mqtt.robot_heartbeat_publisher import (
    RobotHeartbeatPublisher,
)
from isar.services.service_connections.mqtt.robot_info_publisher import (
    RobotInfoPublisher,
)
from isar.services.service_connections.mqtt.robot_status_publisher import (
    RobotStatusPublisher,
)
from isar.state_machine.state_machine import StateMachine, main
from isar.storage.uploader import Uploader
from robot_interface.robot_interface import RobotInterface

if __name__ == "__main__":
    injector: Injector = get_injector()

    keyvault_client = injector.get(Keyvault)
    setup_loggers(keyvault=keyvault_client)
    logger: Logger = logging.getLogger("main")

    state_machine: StateMachine = injector.get(StateMachine)
    uploader: Uploader = injector.get(Uploader)
    robot: RobotInterface = injector.get(RobotInterface)
    queues: Queues = injector.get(Queues)

    threads: List[Thread] = []

    state_machine_thread: Thread = Thread(
        target=main, name="ISAR State Machine", args=[state_machine], daemon=True
    )
    threads.append(state_machine_thread)

    uploader_thread: Thread = Thread(
        target=uploader.run, name="ISAR Uploader", daemon=True
    )
    threads.append(uploader_thread)

    if settings.MQTT_ENABLED:
        mqtt_client: MqttClient = MqttClient(mqtt_queue=queues.mqtt_queue)

        mqtt_thread: Thread = Thread(
            target=mqtt_client.run, name="ISAR MQTT Client", daemon=True
        )
        threads.append(mqtt_thread)

        robot_status_publisher: RobotStatusPublisher = RobotStatusPublisher(
            mqtt_queue=queues.mqtt_queue, robot=robot, state_machine=state_machine
        )
        robot_status_thread: Thread = Thread(
            target=robot_status_publisher.run,
            name="ISAR Robot Status Publisher",
            daemon=True,
        )
        threads.append(robot_status_thread)

        robot_info_publisher: RobotInfoPublisher = RobotInfoPublisher(
            mqtt_queue=queues.mqtt_queue
        )
        robot_info_thread: Thread = Thread(
            target=robot_info_publisher.run,
            name="ISAR Robot Info Publisher",
            daemon=True,
        )
        threads.append(robot_info_thread)

        robot_heartbeat_publisher: RobotHeartbeatPublisher = RobotHeartbeatPublisher(
            mqtt_queue=queues.mqtt_queue
        )

        robot_heartbeat_thread: Thread = Thread(
            target=robot_heartbeat_publisher.run,
            name="ISAR Robot Heartbeat Publisher",
            daemon=True,
        )
        threads.append(robot_heartbeat_thread)

        publishers: List[Thread] = robot.get_telemetry_publishers(
            queue=queues.mqtt_queue,
            robot_name=settings.ROBOT_NAME,
            isar_id=settings.ISAR_ID,
        )
        if publishers:
            threads.extend(publishers)

    api: API = injector.get(API)
    api_thread: Thread = Thread(target=api.run_app, name="ISAR API", daemon=True)
    threads.append(api_thread)

    for thread in threads:
        thread.start()
        logger.info(f"Started thread: {thread.name}")

    while True:
        for thread in threads:
            if not thread.is_alive():
                logger.critical("Thread '%s' failed - ISAR shutting down", thread.name)
                exit(1)
        time.sleep(state_machine.sleep_time)



from isar.apis.models.models import (
    InputOrientation,
    InputPose,
    InputPosition,
    StartMissionResponse,
    StepResponse,
    TaskResponse,
)
from isar.apis.models.start_mission_definition import (
    InspectionTypes,
    StartMissionDefinition,
    StartMissionInspectionDefinition,
    StartMissionTaskDefinition,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import Task
from tests.mocks.step import MockStep


class MockMissionDefinition:
    mock_input_position = InputPosition(x=1, y=1, z=1, frame_name="robot")
    mock_input_orientation = InputOrientation(x=0, y=0, z=0, w=0, frame_name="robot")
    mock_input_pose = InputPose(
        position=mock_input_position,
        orientation=mock_input_orientation,
        frame_name="robot",
    )
    mock_input_target_position = InputPosition(x=5, y=5, z=5, frame_name="robot")
    default_mission = Mission(
        id="default_mission",
        tasks=[
            Task(
                steps=[MockStep.take_image_in_coordinate_direction, MockStep.drive_to]
            ),
            Task(
                steps=[
                    MockStep.take_image_in_coordinate_direction,
                    MockStep.take_image_in_coordinate_direction,
                ]
            ),
        ],
    )
    mock_start_mission_inspection_definition = StartMissionInspectionDefinition(
        type=InspectionTypes.image,
        inspection_target=mock_input_target_position,
    )
    mock_start_mission_inspection_definition_id_123 = StartMissionInspectionDefinition(
        type=InspectionTypes.image,
        inspection_target=mock_input_target_position,
        id="123",
    )
    mock_start_mission_inspection_definition_id_123456 = (
        StartMissionInspectionDefinition(
            type=InspectionTypes.image,
            inspection_target=mock_input_target_position,
            id="123456",
        )
    )
    mock_start_mission_response = StartMissionResponse(
        id=default_mission.id,
        tasks=[
            TaskResponse(
                id=task.id,
                tag_id=task.tag_id,
                steps=[
                    StepResponse(id=step.id, type=step.__class__.__name__)
                    for step in task.steps
                ],
            )
            for task in default_mission.tasks
        ],
    )
    mock_start_mission_definition = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
            ),
        ]
    )
    mock_start_mission_definition_task_ids = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[
                    mock_start_mission_inspection_definition_id_123,
                    mock_start_mission_inspection_definition_id_123456,
                ],
                id="123",
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
                id="123456",
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
                id="123456789",
            ),
        ]
    )
    mock_start_mission_definition_step_ids = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[
                    mock_start_mission_inspection_definition_id_123,
                    mock_start_mission_inspection_definition_id_123456,
                ],
                id="123",
            )
        ]
    )
    mock_start_mission_definition_with_duplicate_task_ids = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
                id="123",
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
                id="123456",
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[mock_start_mission_inspection_definition],
                id="123",
            ),
        ]
    )
    mock_start_mission_definition_with_duplicate_step_ids_cross_tasks = (
        StartMissionDefinition(
            tasks=[
                StartMissionTaskDefinition(
                    pose=mock_input_pose,
                    tag="dummy_tag",
                    inspections=[
                        mock_start_mission_inspection_definition_id_123,
                        mock_start_mission_inspection_definition,
                        mock_start_mission_inspection_definition,
                    ],
                ),
                StartMissionTaskDefinition(
                    pose=mock_input_pose,
                    tag="dummy_tag",
                    inspections=[
                        mock_start_mission_inspection_definition,
                        mock_start_mission_inspection_definition_id_123,
                    ],
                ),
            ]
        )
    )
    mock_start_mission_definition_with_duplicate_step_ids = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspections=[
                    mock_start_mission_inspection_definition_id_123,
                    mock_start_mission_inspection_definition_id_123,
                ],
            ),
        ]
    )



from typing import List, Union

import pytest

from isar.apis.models.start_mission_definition import get_duplicate_ids
from robot_interface.models.mission.step import STEPS, Step
from robot_interface.models.mission.task import Task

task_1: Task = Task([], tag_id=None, id="123")
task_2: Task = Task([], tag_id=None, id="123")
task_3: Task = Task([], tag_id=None, id="123456")
task_4: Task = Task([])
task_5: Task = Task([])

step_1: Step = Step()
step_1.id = "123"
step_2: Step = Step()
step_2.id = "123"
step_3: Step = Step()
step_3.id = "123456"


@pytest.mark.parametrize(
    "item_list, expected_boolean",
    [
        (
            [task_1, task_2, task_3],
            True,
        ),
        (
            [task_1, task_3, task_4, task_5],
            False,
        ),
        (
            [step_1, step_2, step_3],
            True,
        ),
        (
            [step_1, step_3],
            False,
        ),
        (
            [step_1, step_3],
            False,
        ),
    ],
)
def test_duplicate_id_check(
    item_list: Union[List[Task], List[STEPS]], expected_boolean: bool
):
    duplicates: List[str] = get_duplicate_ids(item_list)
    has_duplicates: bool = len(duplicates) > 0
    assert has_duplicates == expected_boolean
